--|
--| Filename        : $Source: /home/ACDE/Work/MySQL/RCS/db.adb,v $
--| Description     : Database interface to MySQL
--| Author          : Michael Erdmann
--| Created On      : 25.3.1999
--| Last Modified By: $Author: ACDE $
--| Last Modified On: $Date: 1999/10/30 15:50:25 $
--| Status          : $State: Exp $
--|
--| Functional Description
--| ======================
--| This components implements a thick binding to the MySQL product
--| which is available for the Linux Plattform.
--|
--|
--| Restrictions
--| ============
--| Only 100 fields per query
--| Number of Queries per connection is 10.
--|
--| References
--| ==========
--|
--|
--| History
--| =======
--| $Log: db.adb,v $
--| Revision 1.5  1999/10/30 15:50:25  ACDE
--| float fields supported
--|
--| Revision 1.4  1999/10/17 15:17:12  ACDE
--| Reading string field does work. Other types still missing
--|
--| Revision 1.3  1999/10/17 09:42:43  ACDE
--| intermediate checkin. The result access has to be reworked completely
--|
--| Revision 1.2  1999/10/16 17:21:13  ACDE
--| Version for component Test. Only String_Field is implemented.
--|
--| Revision 1.1  1999/10/10 17:12:50  ACDE
--| No comments
--|
--|

--* Ada
with System;                  use System;
with System.Storage_Elements; use System.Storage_Elements;
with System.Address_To_Access_Conversions;

with Ada.Exceptions;          use Ada.Exceptions;
with Ada.Strings.Unbounded;   use Ada.Strings.Unbounded;
with Ada.Unchecked_Conversion;
with Ada.Text_IO;             use Ada.Text_IO;
with Ada.Characters.Latin_1;  use Ada.Characters;
with Ada.Calendar;            use Ada.Calendar;
with Ada.Strings.Fixed;       use Ada.Strings.Fixed;

with Interfaces.C;            use Interfaces.C;
use  Interfaces;
with Interfaces.C.Strings;    use Interfaces.C.Strings;

with Unchecked_Deallocation;

package body DB is

   Version : constant String := "$Id: db.adb,v 1.5 1999/10/30 15:50:25 ACDE Exp ACDE $";

   --|
   --| Interface to MYSQL Library. Some of the data types are simply
   --| Addresses because the internals of these strictured do not have 
   --| to be known here.
   --|
   subtype MYSQL_SERVER is  System.Address;
   subtype MYSQL_ROW    is  System.Address;
   subtype MYSQL_RES    is  System.Address;
   subtype MYSQL_FNO    is  C.unsigned;              


   function mysql_connect( mysql : MYSQL_SERVER;
                           host  : Strings.chars_ptr;
                           user  : Strings.chars_ptr;
                           passwd: Strings.chars_ptr)
                                   return MYSQL_SERVER; 
   pragma Import(C, mysql_connect, "mysql_connect");        

   procedure mysql_close(sock: MYSQL_SERVER);             
   pragma Import(C, mysql_close, "mysql_close");  

   function mysql_select_db(mysql: MYSQL_SERVER; db : Strings.chars_ptr) return c.int;       
   pragma Import(C, mysql_select_db, "mysql_select_db");     

   function mysql_query(mysql: MYSQL_SERVER; q : Strings.chars_ptr) return c.int;    
   pragma Import(C, mysql_query, "mysql_query");          

   function mysql_fetch_row(mysql: MYSQL_RES) return MYSQL_ROW;
   pragma Import(C, mysql_fetch_row, "mysql_fetch_row");   

   function mysql_field_seek(mysql : MYSQL_RES; offset: MYSQL_FNO) return MYSQL_FNO;
   pragma Import(C, mysql_field_seek, "mysql_field_seek");    -- mysql.h:166

   type MYSQL_field_types is (                                  -- mysql_com.h:64
        FIELD_TYPE_DECIMAL,                                     -- mysql_com.h:64
        FIELD_TYPE_TINY,                                        -- mysql_com.h:64
        FIELD_TYPE_SHORT,                                       -- mysql_com.h:65
        FIELD_TYPE_LONG,                                        -- mysql_com.h:65
        FIELD_TYPE_FLOAT,                                       -- mysql_com.h:66
        FIELD_TYPE_DOUBLE,                                      -- mysql_com.h:66
        FIELD_TYPE_NULL,                                        -- mysql_com.h:67
        FIELD_TYPE_TIMESTAMP,                                   -- mysql_com.h:67
        FIELD_TYPE_LONGLONG,                                    -- mysql_com.h:68
        FIELD_TYPE_INT24,                                       -- mysql_com.h:68
        FIELD_TYPE_DATE,                                        -- mysql_com.h:69
        FIELD_TYPE_TIME,                                        -- mysql_com.h:69
        FIELD_TYPE_DATETIME,                                    -- mysql_com.h:70
        FIELD_TYPE_YEAR,                                        -- mysql_com.h:70
        FIELD_TYPE_NEWDATE,                                     -- mysql_com.h:71
        FIELD_TYPE_ENUM,                                        -- mysql_com.h:72
        FIELD_TYPE_SET,                                         -- mysql_com.h:73
        FIELD_TYPE_TINY_BLOB,                                   -- mysql_com.h:74
        FIELD_TYPE_MEDIUM_BLOB,                                 -- mysql_com.h:75
        FIELD_TYPE_LONG_BLOB,                                   -- mysql_com.h:76
        FIELD_TYPE_BLOB,                                        -- mysql_com.h:77
        FIELD_TYPE_VAR_STRING,                                  -- mysql_com.h:78
        FIELD_TYPE_STRING                                       -- mysql_com.h:80
    );
    for MYSQL_field_types use (                             -- mysql_com.h:64
        FIELD_TYPE_DECIMAL => 0,                                -- mysql_com.h:64
        FIELD_TYPE_TINY => 1,                                   -- mysql_com.h:64
        FIELD_TYPE_SHORT => 2,                                  -- mysql_com.h:65
        FIELD_TYPE_LONG => 3,                                   -- mysql_com.h:65
        FIELD_TYPE_FLOAT => 4,                                  -- mysql_com.h:66
        FIELD_TYPE_DOUBLE => 5,                                 -- mysql_com.h:66
        FIELD_TYPE_NULL => 6,                                   -- mysql_com.h:67
        FIELD_TYPE_TIMESTAMP => 7,                              -- mysql_com.h:67
        FIELD_TYPE_LONGLONG => 8,                               -- mysql_com.h:68
        FIELD_TYPE_INT24 => 9,                                  -- mysql_com.h:68
        FIELD_TYPE_DATE => 10,                                  -- mysql_com.h:69
        FIELD_TYPE_TIME => 11,                                  -- mysql_com.h:69
        FIELD_TYPE_DATETIME => 12,                              -- mysql_com.h:70
        FIELD_TYPE_YEAR => 13,                                  -- mysql_com.h:70
        FIELD_TYPE_NEWDATE => 14,                               -- mysql_com.h:71
        FIELD_TYPE_ENUM => 247,                                 -- mysql_com.h:72
        FIELD_TYPE_SET => 248,                                  -- mysql_com.h:73
        FIELD_TYPE_TINY_BLOB => 249,                            -- mysql_com.h:74
        FIELD_TYPE_MEDIUM_BLOB => 250,                          -- mysql_com.h:75
        FIELD_TYPE_LONG_BLOB => 251,                            -- mysql_com.h:76
        FIELD_TYPE_BLOB => 252,                                 -- mysql_com.h:77
        FIELD_TYPE_VAR_STRING => 253,                           -- mysql_com.h:78
        FIELD_TYPE_STRING => 254                                -- mysql_com.h:80
    );
   for MYSQL_field_types'size use 32;          

   type st_mysql_field is                                       -- mysql.h:39
        record
            name      : Strings.chars_ptr;                      -- mysql.h:40
            table     : Strings.chars_ptr;                      -- mysql.h:41
            def       : Strings.chars_ptr;                      -- mysql.h:42
            c_type    : MYSQL_field_types;                      -- mysql.h:43
            length    : c.unsigned;                             -- mysql.h:44
            max_length: c.unsigned;                             -- mysql.h:45
            flags     : c.unsigned;                             -- mysql.h:46
            decimals  : c.unsigned;                             -- mysql.h:47
        end record;

   pragma Convention(C, st_mysql_field);                        -- mysql.h:39
   type MYSQL_Field is access all st_mysql_field;


   function mysql_fetch_field(handle: MYSQL_RES) return MYSQL_FIELD;
   pragma Import(C, mysql_fetch_field, "mysql_fetch_field");

   function mysql_store_result(mysql: MYSQL_SERVER) return MYSQL_RES; -- mysql.h:161
   pragma Import(C, mysql_store_result, "mysql_store_result"); -- mysql.h:161

   function mysql_use_result(mysql: MYSQL_SERVER) return MYSQL_RES;   -- mysql.h:162
   pragma Import(C, mysql_use_result, "mysql_use_result");     -- mysql.h:162

   procedure mysql_free_result(result: MYSQL_RES);             -- mysql.h:163
   pragma Import(C, mysql_free_result, "mysql_free_result");   
 
   ---====================================================================---
   ---===                O B J  E C T     D A T A                      ===---
   ---====================================================================---

   type Field_Info_Array is array( Field_Number ) of St_MySql_Field;

   --|
   --| This is the instance of the component internal data.
   --|
   type Query_Context is record
         In_Use : Boolean   := False;
         Result : MYSQL_RES := Null_Address;
	 Row    : MYSQL_ROW := Null_Address;
	 Field  : Field_Info_Array;
      end record;

   type Query_Table is array( Query_ID ) of Query_Context;

   type Object_Data is record
         Server_Name : Unbounded_String;
	 Login_Name  : Unbounded_String;

	 Connected   : Boolean      := False;
         Server      : MYSQL_Server := Null_Address;
	 Query       : Query_Table;
	 
      end record;
   ---=====================================================================---
   ---===         L O C A L   S U P P O R T   P R O C E D U R E S       ===---
   ---=====================================================================---

   --|
   --| Retrieve the address from an array fo addresses
   --|
   function Get_Address( row : MYSQL_ROW; pos : Field_Number ) return Address is
      offset : Storage_Offset := 
                   Storage_Offset( (Pos-1) * Address'Size/Storage_Element'Size );

      package Ptr_To_Address is 
                new System.Address_To_Access_Conversions( Object => Address );
      use Ptr_To_Address;
   begin
      return To_Pointer( row + offset ).all ;
   end Get_Address;

   --|
   --| Copy a string from the memory address. This code is not nice but
   --| is works.
   --|
   function Copy_String ( addr : Address; len : Positive ) return String is
      result : String( 1..len ) := ( others => ' ' );
      byte   : Character;
 
      package Byte_String is 
                    new System.Address_To_Access_Conversions ( Object => Character );
      use Byte_String;

   begin
      for i in 1..len loop
         byte := To_Pointer( addr + Storage_Offset( i-1 ) ).all;
	 if byte /= Latin_1.NUL then
            result(i) := byte;
	 else
	    exit;
	 end if;
      end loop;

      return result;
   end Copy_String;

   --|
   --| Verify if data and query are valid informations.
   --|
   procedure Validate( data : in Object_Data_Access; id : in Query_ID ) is
   begin
      if data = null then
         raise Not_Initialized;
      end if;
    
      if not Data.Query(id).In_Use then
         raise Usage_Error;
      end if;

   end Validate;

   ---======================================================================---
   ---===             C O M P O  N E N T    I N T E R F A C E            ===---
   ---======================================================================---

   --|
   --| Intialize the Object data and call the Initalization
   --| procedure of the extention.
   --|
   procedure Initialize( this  : in out Object'Class ) is
   begin
      if this.data /= null then
         raise Usage_Error;
      end if;

      this.data := new Object_Data;

   end Initialize;

   --|
   --| Finalize the instance by releasing the object resources
   --|
   procedure Finalize( this : in out Object'Class ) is
      data   : Object_Data_Access renames this.data;

      procedure Free is
            new Unchecked_Deallocation( Object_Data, Object_Data_Access);
   begin
      if data = null then
         raise Not_Initialized;
      end if;

      if data.connected = true  then
         Disconnect( this );
      end if;

      for i in Query_ID loop
         if data.query(i).result /= Null_Address then
	    Drop_Query( this, i );
	 end if;
      end loop;

      Free( data );

      this.data := null;

   end Finalize;

   ---=====================================================================---
   ---===           A T T R I B U T E    F U N C T I O N S              ===---
   ---=====================================================================---


   --|
   --| Set the user Name for a server connection
   --|
   procedure User( this : in out Object'Class; Name : in String ) is
      data : Object_Data_Access := this.data;
   begin
      if data = null then
         raise Usage_Error;
      end if;
      
      data.Login_Name := To_Unbounded_String( Name );
   end User;


   ---=====================================================================---
   ---===                        M E T H O D S                          ===---
   ---=====================================================================---

   --|
   --| Connect to the server specified by the Object.
   --|
   procedure Connect( this : in out Object'Class; Server : in String ) is
      data : Object_Data_Access renames this.data;
   begin
      if data = null then
         raise Not_Initialized;
      end if;

      data.Server_Name := To_Unbounded_String( Server );
      data.Server := MYSQL_Connect( Null_Address, 
   		 New_String( To_String( data.Server_Name) ),
		 New_String( To_String(data.Login_Name) ),
		 New_String("") );

      if data.Server = Null_Address then
         raise Connect_Failure;
      end if;

      data.Connected := True;

   end Connect;

   --|
   --| Disconnect from a server.
   --|
   procedure Disconnect( this : in out Object'Class ) is
      data : Object_Data_Access renames this.data;
   begin
      if data = null then
         raise Not_Initialized;
      end if;

      mysql_close( data.Server );
      data.Connected := False;

   end Disconnect;

   --|
   --| Select a data base from the server.
   --|
   procedure Select_DB( this : in out Object'Class; name : in String ) is
      data : Object_Data_Access renames this.data;
      RC   : C.Int := 0;
   begin
      if data = null then
         raise Not_Initialized;
      end if;

      RC :=  mysql_select_db( data.Server, New_String( name ));
      if not ( RC = 0 ) then
         raise DB_Name_Failure;
      end if;
 
   end Select_DB;

   --|
   --| Perform a query on the server and retrieve the result
   --| context. The function returns the id of the query context. 
   --|
   function Query( this : in Object'Class; query : in String ) return Query_ID 
   is
      data       : Object_Data_Access renames this.data;
      Result     : Query_ID := 1;
      RC         : C.int    := 0;
      Slot_Found : Boolean := False;
   begin
      if data = null then
         raise Not_Initialized;
      end if;

      -- perform the query
      RC := mysql_query( data.Server, New_String( query ));
      if RC /= 0 then
         raise No_Query_Result;
      end if;

      -- get a result identifier and the field information
      for i in Query_ID loop
         if data.Query(i).In_Use = False then
	    Result := i;

	    data.Query(i).In_Use := True;
            data.Query(i).Result := mysql_store_result( data.Server );
            data.Query(i).Row    := mysql_fetch_row(data.Query(i).Result);

            for i in Field_Number loop 
               declare
                  Field : MYSQL_Field := mysql_fetch_field( Data.Query(Result).Result  );
	       begin
	         exit when Field = null;
	    
	         data.Query( Result ).Field(i) := Field.all;
	       end;
            end loop;

	    Slot_Found := True;
	    exit;

	 end if;
      end loop;

      if Slot_Found then
         return Result;
      else
         raise Out_Of_Resources;
      end if;

   end Query;

   --|
   --| Release the query context
   --|
   procedure Drop_Query( this : in out Object'Class; id : in Query_ID ) is
      data : Object_Data_Access renames this.data;
   begin
      Validate( data, id );

      if data.Query(id).In_Use then
         mysql_free_result( data.Query(id).Result);
	 data.Query(id).Result := Null_Address;
	 data.Query(id).In_Use := False;
      else
         raise Usage_Error;
      end if;
   end Drop_Query;

   --|
   --| Get the next matching tuple for the given query
   --|
   procedure Next( this : in out Object'Class; query : in Query_ID ) is
      data : Object_Data_Access renames this.data;
   begin
      Validate( data, query );

      if data.Query(query).Result /= Null_Address then
         data.Query(query).Row := mysql_fetch_row( data.Query(query).Result );
	 if Data.Query(query).Row = Null_Address then
	    raise No_More_Rows;
	 end if;
      else
         raise Usage_Error;
      end if;
   end Next;

   --|
   --| Translaste the name into a field number.
   --|
   function Get_Field_Number( this : Object; id : Query_ID; name : String) 
                                                                    return Field_Number is
      data   : Object_Data_Access renames this.data;      
   begin
      Validate( data, id );

      for i in Field_Number loop
         if Value( Data.Query(id).Field(i).Name ) = name then
	    return i;
	 end if;
      end loop; 

      raise Unknown_Field;
   end Get_Field_Number;

   --|
   --| Get String field identifird by the field number
   --|
   function String_Field( this : in Object; id : in Query_ID; fNumber : in Field_Number )   
                                                                     return Unbounded_String 
   is                          
      data    : Object_Data_Access renames this.data;      
      length  : Positive;
   begin
      Validate( data, id );

      length  := Positive( Data.Query(id).Field(fNumber).length ); 
     
      if data.Query(id).Field(fNumber).c_type = FIELD_TYPE_STRING then
         return To_Unbounded_String(
	                    Copy_String( Get_Address( data.Query(id).Row, fNumber ), length ));
      else
         raise Field_Type_Error; 
      end if ;

   end String_Field;

   --|
   --| Get the named string field
   --|
   function String_Field( this : in Object; id : in Query_ID; name : in String )   
                                                                     return Unbounded_String 
   is                          
      data    : Object_Data_Access renames this.data;      
      fNumber : Field_Number; 
   begin
      fNumber := Get_Field_Number( this, id, name );

      return String_Field( this, id, fNumber );

   end String_Field;

   --|
   --| Get String field identifird by the field number
   --|
   function Date_Field( this : in Object; id : in Query_ID; fNumber : in Field_Number )   
                                                                     return Time 
   is                          
      data    : Object_Data_Access renames this.data;      
      length  : Positive;
   begin
      Validate( data, id );
  
      length  := Positive( Data.Query(id).Field(fNumber).length ); 
     
      if data.Query(id).Field(fNumber).c_type = FIELD_TYPE_DATE then
         declare
	    Time_String : String(1..length);
	    Year        : Year_Number;
	    Month       : Month_Number;
	    Day         : Day_Number;
	    tmp         : String(1..5);

	    next        : Positive := 1;

	    -- copy a date element into the tmp string and retirn the start offset
	    procedure Copy_Element( Str : String ) is
	    begin
	       tmp := (others => ' ');

	       for i in 1..tmp'Length loop
	          if Str(next) = '-' then
		     next := next + 1;
		     exit;
		  end if;
		  tmp(i) := Str(next);	

	          exit when next = length;
		  next := next + 1;
	       end loop;

	    exception 
	       when Others =>
	          put_line("Error in Copy_Element : " & Str & " at " & Integer'Image(next) );
		  raise;

	    end Copy_Element;

	 begin
            Move( Source => Copy_String( Get_Address( data.Query(id).Row, fNumber ), length ),
	          Target => Time_String);

	    Copy_Element( Time_String );
	    Year  := Year_Number'Value( tmp );


	    Copy_Element( Time_String );
	    Month := Month_Number'Value( tmp );
 
	    Copy_Element( Time_String );
	    Day   := Day_Number'Value( tmp );
	    	    
	    return Time_Of( Year, Month, Day );
	     
	 exception
	    when The_Error : Others =>
	        raise Field_Parse_Error;
	 end;
      else
         raise Field_Type_Error; 
      end if;

   end Date_Field;

   --|
   --| Get the named string field
   --|
   function Date_Field( this : in Object; id : in Query_ID; name : in String ) return Time 
   is                          
      data    : Object_Data_Access renames this.data;      
      fNumber : Field_Number; 
   begin    
      fNumber := Get_Field_Number( this, id, name );

      return Date_Field( this, id, fNumber );

   end Date_Field;

   --|
   --| return the blob field given by the columne number
   --|
   function Blob_Field ( this : in Object; id : in Query_ID; fNumber : in  Field_Number ) 
                                                                   return Blob is
      data   : Object_Data_Access renames this.data;
      length : Positive;
   begin
      Validate( data, id );
      			 
      length  := Positive( Data.Query(id).Field(fNumber).length ); 
     
      if data.Query(id).Field(fNumber).c_type = FIELD_TYPE_BLOB or 
         data.Query(id).Field(fNumber).c_type = FIELD_TYPE_TINY_BLOB or 
         data.Query(id).Field(fNumber).c_type = FIELD_TYPE_MEDIUM_BLOB or              
         data.Query(id).Field(fNumber).c_type = FIELD_TYPE_LONG_BLOB or  
         data.Query(id).Field(fNumber).c_type = FIELD_TYPE_BLOB 
      then
         return Copy_String( Get_Address( data.Query(id).Row, fNumber ), length );
      else
         raise Field_Type_Error;
      end if;

   end Blob_Field;

   --|
   --| return the named blob field
   --|
   function Blob_Field( this : in Object; id : in Query_ID; name : in String ) return Blob 
   is                          
      data    : Object_Data_Access renames this.data;      
      fNumber : Field_Number; 
   begin
      fNumber := Get_Field_Number( this, id, name );

      return Blob_Field( this, id, fNumber );

   end Blob_Field;

   --|
   --|  return a number field as an integer.
   --|
   function Integer_Field( this : in Object; id : in Query_ID;  fNumber : in Field_Number ) 
                                                     return Integer is
      data   : Object_Data_Access renames this.data;
      length : Positive;
   begin
      Validate( data, id );
      			 
      length  := Positive( Data.Query(id).Field(fNumber).length ); 
    
      if data.Query(id).Field(fNumber).c_type = FIELD_TYPE_TINY or                                    -- mysql_com.h:64
         data.Query(id).Field(fNumber).c_type = FIELD_TYPE_SHORT or
         data.Query(id).Field(fNumber).c_type = FIELD_TYPE_LONG
      then 
         return Integer'Value(Copy_String( Get_Address(data.Query(id).Row, fNumber),length));
      else
         raise Field_Type_Error;
      end if;
   end Integer_Field;

   --|
   --| return the number field by name
   --|
   function Integer_Field( this : in Object; id : in Query_ID; name : in String ) return Integer
   is                          
      data    : Object_Data_Access renames this.data;      
      fNumber : Field_Number; 
   begin    
      fNumber := Get_Field_Number( this, id, name );

      return Integer_Field( this, id, fNumber );

   end Integer_Field;

   --|
   --| return a float field
   --|
   function Float_Field( this : in Object; id : in Query_ID;  fNumber : in Field_Number ) 
                                                     return Float is
      data   : Object_Data_Access renames this.data;
      length : Positive;
   begin
      Validate( data, id );
      			 
      length  := Positive( Data.Query(id).Field(fNumber).length ); 
    
      if data.Query(id).Field(fNumber).c_type = FIELD_TYPE_FLOAT or                                    -- mysql_com.h:64
         data.Query(id).Field(fNumber).c_type = FIELD_TYPE_DOUBLE 
      then 
         return Float'Value(Copy_String( Get_Address(data.Query(id).Row, fNumber),length));
      else
         raise Field_Type_Error;
      end if;
   end Float_Field;

   --|
   --| return the number field by name
   --|
   function Float_Field( this : in Object; id : in Query_ID; name : in String ) return Float
   is                          
      data    : Object_Data_Access renames this.data;      
      fNumber : Field_Number; 
   begin
      Validate( data, id );
    
      fNumber := Get_Field_Number( this, id, name );

      return Float_Field( this, id, fNumber );

   end Float_Field;


end DB;

